# SPDX-License-Identifier: AGPL-3.0-or-later OR GPL-2.0-or-later OR CERN-OHL-S-2.0+ OR Apache-2.0
import abc
from itertools import combinations
from typing import (
    Any, List, Tuple, Dict, Union, Iterable, Optional, overload,
)

from pdkmaster import _util
from pdkmaster.technology import geometry as _geo, mask as _msk
from pdkmaster.design import layout as _lay, cell as _cell, library as _lbry
from pdkmaster.design.layout import layout_ as _laylay
from .export import export2db

import pya

__all__ = ["merge"]


def _export_polygon(polygon: _geo.Polygon) -> pya.Region:
    """Convert PKDMaster _Shape to klayout db object

    The object is converted to integer based shape in order to insert it in
    a klayout Region object

    API Notes:
        This function is for internal use and does not have backwards compatibility
        guarantee.
    """
    dshape = export2db(polygon)
    return pya.Region(dshape.to_itype(_geo.epsilon))


def _import_polygon(polygon) -> _geo.Polygon:
    """Convert klayout db to PKDMaster Polygon object

    API Notes:
        This function is for internal use and does not have backwards compatibility
        guarantee.
    """
    if isinstance(polygon, pya.Box):
        box = polygon.to_dtype(_geo.epsilon)
        return _geo.Rect(
            left=box.left, bottom=box.bottom, right=box.right, top=box.top,
        )
    elif polygon.is_box():
        box = polygon.bbox().to_dtype(_geo.epsilon)
        return _geo.Rect(
            left=box.left, bottom=box.bottom, right=box.right, top=box.top,
        )
    else:
        poly2 = (
            polygon if isinstance(polygon, pya.SimplePolygon)
            else polygon.to_simple_polygon()
        )
        spoly = poly2.to_dtype(_geo.epsilon)
        p0 = _util.get_first_of(spoly.each_point())
        return _geo.Polygon(points=(
            *(_geo.Point(x=p.x, y=p.y) for p in spoly.each_point()),
            _geo.Point(x=p0.x, y=p0.y),
        ))


def _import_regionshape(region: pya.Region) -> _geo.Polygon:
    """Convert pya.Region containing a single polygon to a PDKMaster _geo.Polygon object.
    """
    region.merge()

    polys = tuple(region.each())
    assert len(polys) == 1

    return _import_polygon(polys[0])


class _MPSDictElemMPS:
    def __init__(self, *, mps_orig: _geo.MultiPartShape):
        partshapes = list(part.partshape for part in mps_orig.parts)
        self.partregions = list(
            _export_polygon(partshape) for partshape in partshapes
        )
        # Parts can be removed, mark them by deferring original number to new number
        self.partidcs: List[int] = list(range(len(mps_orig.parts)))
        # Only create mps the first time a dereference is done.
        # Don't allow to marge shapes anymore after the new mps has been created.
        self.mps: Optional[_geo.MultiPartShape] = None
        self.mpsidcs: Optional[Tuple[int, ...]] = None

    @property
    def is_dereffed(self) -> bool:
        return self.mps is not None

    def lookup_partidx(self, idx: int):
        while True:
            idx2 = self.partidcs[idx]
            if idx == idx2:
                return idx
            else:
                idx = idx2

    def part_is_merged(self, idx: int)  -> bool:
        return self.partidcs[idx] != idx

    def deref_part(self, idx: int):
        if self.mps is None:
            lookedup = list(self.lookup_partidx(i) for i in range(len(self.partregions)))
            uni = list(sorted(set(lookedup)))
            partregions = tuple(self.partregions[idx] for idx in uni)
            parts = map(_import_regionshape, partregions)
            fullshaperegion = sum(partregions, pya.Region())
            fullshape = _import_regionshape(fullshaperegion)
            self.mps = _geo.MultiPartShape(fullshape=fullshape, parts=parts)
            self.mpsidcs = tuple(uni.index(i) for i in lookedup)
        assert self.mpsidcs is not None
        return self.mps.parts[self.mpsidcs[idx]]

    def change_shape(self, *, idx: int, shaperegion: pya.Region):
        assert not self.is_dereffed

        idx2 = self.lookup_partidx(idx)
        self.partregions[idx2] = shaperegion

    def __hash__(self) -> int:
        raise TypeError(f"{self.__class__.__name__} objects are mutable and can't be hashed")

    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other)


class _MPSDictElem:
    """Class for the values stored in _MPSDict

    It stores extra info needed during conversion of all MultiPartShape._Part objects.

    API Notes:
        This class is for internal use and does not have backwards compatibility
        guarantee.
    """
    def __init__(self, *,
        mps: _geo.MultiPartShape, mpsdict: "_MPSDict",
    ):
        self.mps_orig = mps
        self.elemmps: Optional[_MPSDictElemMPS]
        self.elemmps = None
        self.partidcs: Optional[List[int]]
        self.partidcs = None
        self.mpsdict = mpsdict
        fullshaperegion = _export_polygon(mps.fullshape)
        self._partregions = partregions = list(
            _export_polygon(part.partshape) for part in mps.parts
        )

        # Check sum of parts is fullshape
        region = sum(partregions, pya.Region())
        region.merge()
        if not (region ^ fullshaperegion).is_empty():
            raise ValueError("MultiPartShape parts do not match fullshape")

        # Check that parts don't overlap
        for part1, part2 in combinations(self.partregions, 2):
            overlaps = not part1.overlapping(part2).is_empty()
            if overlaps:
                raise ValueError(
                    "Overlapping parts in MultiPartShape object",
                )

    def init_elemmps(self):
        assert self.elemmps is None, "Internal error"
        self.elemmps = _MPSDictElemMPS(mps_orig=self.mps_orig)
        self.partidcs = list(range(len(self.mps_orig.parts)))

    def merge_with(self, other: "_MPSDictElem") -> int:
        """
        Return:
            the index offset to add to the index value of other
        """
        if self.elemmps is None:
            self.init_elemmps()
        assert self.elemmps is not None
        if other.elemmps is None:
            other.init_elemmps()
        assert (other.elemmps is not None) and (other.partidcs is not None)

        if self.elemmps == other.elemmps:
            return 0
        else:
            offset = len(self.elemmps.partregions)
            elemmps2 = other.elemmps
            self.elemmps.partregions.extend(elemmps2.partregions)
            self.elemmps.partidcs.extend(idx + offset for idx in elemmps2.partidcs)
            # Replace all accorences of old elemmps with new one
            for elem in self.mpsdict.values():
                if (elem != self) and (elem.elemmps == elemmps2):
                    assert elem.partidcs is not None
                    assert elem.elemmps is not None
                    elem.elemmps = self.elemmps
                    elem.partidcs = list(idx + offset for idx in elem.partidcs)
            return offset

    @property
    def partregions(self) -> List[pya.Region]:
        elemmps = self.elemmps
        if elemmps is None:
            return self._partregions
        else:
            assert self.partidcs is not None
            return list(
                elemmps.partregions[elemmps.lookup_partidx(idx)]
                for idx in self.partidcs
            )

    def change_shape_part(self, *, idx: int, shape: _geo.Polygon):
        shaperegion = _export_polygon(shape)
        self._partregions[idx] = shaperegion

        assert (self.elemmps is not None) and (self.partidcs is not None)
        self.elemmps.change_shape(idx=self.partidcs[idx], shaperegion=shaperegion)

    def part_is_merged(self, *, idx: int) -> bool:
        if self.elemmps is None:
            return False
        else:
            assert self.partidcs is not None
            return self.elemmps.part_is_merged(idx=self.partidcs[idx])

    def deref_part(self, orig_idx: int) -> _geo.MultiPartShape._Part:
        if self.elemmps is None:
            return self.mps_orig.parts[orig_idx]
        else:
            assert self.partidcs is not None
            return self.elemmps.deref_part(self.partidcs[orig_idx])

    def __hash__(self) -> int:
        raise TypeError(f"{self.__class__.__name__} objects are mutable and can't be hashed")

    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other)
        # assert False, "Internal error" # pragma: no cover


class _Reffed:
    @abc.abstractmethod
    def __init__(self):
        return # pragma: no cover

    @abc.abstractmethod
    def deref(self) -> Union[_geo._Shape, _geo.MaskShape, _geo.MaskShapes, _laylay._SubLayout]:
        raise RuntimeError("Non overloaded abstract method") # pragma: no cover


class _ShapeReffed(_Reffed):
    @abc.abstractmethod
    def deref(self) -> _geo._Shape:
        raise RuntimeError("Non overloaded abstract method") # pragma: no cover


class _ShapeRef(_ShapeReffed):
    def __init__(self, *, shape: _geo._Shape):
        self.shape = shape

    def deref(self) -> _geo._Shape:
        return self.shape


class _PartRef(_ShapeReffed):
    def __init__(self, *, elem: _MPSDictElem, idx: int):
        self.elem = elem
        self.idx = idx

    @property
    def is_merged(self):
        return self.elem.part_is_merged(idx=self.idx)
    @property
    def region(self):
        return _util.get_nth_of(self.elem.partregions, n=self.idx)

    def deref(self) -> _geo.MultiPartShape._Part:
        return self.elem.deref_part(self.idx)

    def interacts_with(self, *, shapereg) -> bool:
        return not self.region.interacting(shapereg).is_empty()

    def is_same_part_as(self, other: "_PartRef") -> bool:
        elem = self.elem
        elem2 = other.elem
        if elem.elemmps is None:
            return (
                (elem2.elemmps is None)
                and (elem.mps_orig == elem2.mps_orig)
                and (self.idx == other.idx)
            )
        else:
            elemmps = elem.elemmps
            elemmps2 = elem2.elemmps
            return (
                (elemmps == elemmps2)
                and (elemmps2 is not None)
                and (elem.partidcs is not None)
                and (elem2.partidcs is not None)
                and (
                    elemmps.lookup_partidx(idx=elem.partidcs[self.idx])
                    == elemmps2.lookup_partidx(idx=elem2.partidcs[other.idx])
                )
            )

    def add_polygon(self, polygon: pya.Polygon) -> None:
        elem = self.elem
        if elem.elemmps is None:
            elem.init_elemmps()
        assert elem.elemmps is not None

        region = elem.partregions[self.idx]
        region += polygon # type: ignore

        self.change_shape_to(_import_regionshape(region))

    def change_shape_to(self, shape: _geo.Polygon) -> None:
        elem = self.elem
        elem.change_shape_part(idx=self.idx, shape=shape)

    def merge_with(self, ref: "_PartRef") -> None:
        elem = self.elem
        elem2 = ref.elem
        if (elem.elemmps is None) or (elem.elemmps != elem2.elemmps):
            elem.merge_with(elem2)
            # The merge_with call above will also ensure elem.init_elemmps() has been
            # called
        elemmps = elem.elemmps
        elem2mps = elem2.elemmps
        assert (
            (elemmps is not None) and (elemmps == elem2mps)
            and (elem.partidcs is not None) and (elem2.partidcs is not None)
            and (elemmps.partidcs is not None)
        )

        # Merge other shape in our shape
        region = self.region
        region += ref.region
        self.change_shape_to(_import_regionshape(region))

        # Mark second part as merged to the first one
        partidx = elem.partidcs[self.idx]
        partidx2 = elem2.partidcs[ref.idx]
        elemmps.partidcs[partidx2] = partidx


class _MultiShapeRef(_ShapeReffed):
    def __init__(self, *, shapes: Iterable[Union[_ShapeRef, _PartRef]]):
        self.shapes = shapes

    def deref(self) -> _geo._Shape:
        # Multiple shape can point to the same part and end up with one shape
        shapes = tuple(shape.deref() for shape in self.shapes)
        return _geo.MultiShape(shapes=shapes)


class _MaskShapeRef(_Reffed):
    def __init__(self, *, mask: _msk.DesignMask, ref: _ShapeReffed):
        self.mask = mask
        self.ref = ref

    def deref(self) -> _geo.MaskShape:
        return _geo.MaskShape(mask=self.mask, shape=self.ref.deref())


class _MaskShapesRef(_Reffed):
    def __init__(self, *, maskshapes: Iterable[_MaskShapeRef]):
        self.maskshapes = tuple(maskshapes)

    def deref(self) -> _geo.MaskShapes:
        return _geo.MaskShapes(ms.deref() for ms in self.maskshapes)


class _SubLayoutReffed(_Reffed):
    @abc.abstractmethod
    def deref(self) -> _laylay._SubLayout:
        raise RuntimeError("Non overloaded abstract method") # pragma: no cover


class _SubLayoutRef(_SubLayoutReffed):
    def __init__(self, *, sublayout: _laylay._SubLayout):
        self.sublayout = sublayout

    def deref(self) -> _laylay._SubLayout:
        return self.sublayout


class _MaskShapesSubLayoutRef(_SubLayoutReffed):
    def __init__(self, *, net, ref: _MaskShapesRef):
        self.net = net
        self.ref = ref

    def deref(self) -> _laylay._MaskShapesSubLayout:
        return _laylay._MaskShapesSubLayout(net=self.net, shapes=self.ref.deref())


class _MPSDict(Dict[_geo.MultiPartShape, _MPSDictElem]):
    # Dict with automatically creation of element
    def __getitem__(self, mps: _geo.MultiPartShape) -> _MPSDictElem:
        try:
            return super().__getitem__(mps)
        except KeyError:
            mps2 = _MPSDictElem(mps=mps, mpsdict=self)
            self[mps] = mps2
            return mps2

    def partref(self, part: _geo.MultiPartShape._Part) -> _PartRef:
        mps = part.multipartshape
        e = self[mps]
        return _PartRef(elem=e, idx=mps.parts.index(part))


class _ShapeMerger:
    def __init__(self):
        self._mps_lookup: _MPSDict
        self._mps_lookup = _MPSDict()

    @property
    def mps_lookup(self) -> _MPSDict:
        return self._mps_lookup

    @overload
    def __call__(self, shape: _lay.LayoutT) -> _lay.LayoutT:
        ... # pragma: no cover
    @overload
    def __call__(self, shape: _geo.MaskShape) -> _MaskShapeRef:
        ... # pragma: no cover
    @overload
    def __call__(self, shape: _geo.MaskShapes) -> _MaskShapesRef:
        ... # pragma: no cover
    @overload
    def __call__(self, shape: _geo.MultiPartShape._Part) -> _PartRef:
        ... # pragma: no cover
    @overload
    def __call__(self, shape: _geo._Shape) -> _ShapeReffed:
        ... # pragma: no cover
    @overload
    def __call__(self, shape: Iterable[_geo._Shape]) -> Tuple[_ShapeReffed, ...]:
        ... # pragma: no cover
    def __call__(self, shape):
        if isinstance(shape, _lay.LayoutT):
            return self.merge_layout(shape)
        elif isinstance(shape, _geo.MaskShape):
            return self.merge_maskshape(shape)
        elif isinstance(shape, _geo.MaskShapes):
            return self.merge_maskshapes(shape)
        elif isinstance(shape, _geo.MultiShape):
            return self.merge_multishape(shape)
        elif isinstance(shape, _geo.MultiPartShape._Part):
            return self._mps_lookup.partref(shape)
        elif isinstance(shape, _geo._Shape):
            return _ShapeRef(shape=shape)
        elif _util.is_iterable(shape):
            # This is not MaskShapes so assume it's Iterable[_geo._Shape]
            return self.merge_shapes(shape)
        else:
            raise TypeError(f"Unsupported shape object type {type(shape)}")

    def merge_maskshape(self, ms: _geo.MaskShape) -> _MaskShapeRef:
        return _MaskShapeRef(mask=ms.mask, ref=self(ms.shape))

    def merge_maskshapes(self, mss: _geo.MaskShapes) -> _MaskShapesRef:
        return _MaskShapesRef(maskshapes=map(self.merge_maskshape, mss))

    def merge_multishape(self, ms: _geo.MultiShape) -> _ShapeReffed:
        """merge the MultiShape.

        This method will use KLayout to merge a much as possible the shapes
        inside a MultiShape object.

        The resulting merged shape. If original MultiShape object only contained
        interacting polygons a _Shape object will be returned that is not a
        MultiShape object.

        As side effect the MultiPartShape objects will also be checked that they
        are properly defined; e.g. that the fullshape is exactly the or-ing of the
        parts without any of the parts overlapping.
        """
        # Sort the shapes
        shapes: List[Union[_ShapeRef, _PartRef]] = [] # Final list of shapes
        partmps_list: List[_PartRef] = []
        polygons = pya.Region()
        # First stage:
        # - add non-merging shapes directly,
        # - merge polygons inside a klayout region,
        # - build list of multishape part objects
        for shape in ms.shapes:
            try:
                area = shape.area
            except:
                # Try to merge on object that don't have area computation implemented
                pass
            else:
                if area < (_geo.epsilon**2) or isinstance(shape, _geo.RepeatedShape):
                    # Zero area shapes are retained as is.
                    shapes.append(_ShapeRef(shape=shape))
                    self(shape)
                    continue

            # MultiShape should not be hierarchical
            assert (not isinstance(shape, _geo.MultiShape)), "Internal Error"
            # MultiPartShape should not be part of MultiShape,
            # (MultiPartShape._Part should)
            assert (not isinstance(shape, _geo.MultiPartShape)), "Internal Error"

            if isinstance(shape, _geo.MultiPartShape._Part):
                mps = shape.multipartshape
                klmps = self.mps_lookup[shape.multipartshape]
                idx = mps.parts.index(shape)
                partmps_list.append(_PartRef(elem=klmps, idx=idx))
            else:
                assert isinstance(shape, _geo.Polygon)
                polygons += _export_polygon(shape)
        polygons.merge()

        # Join polygons from polygons into the MultiPartShape objects
        merged_polygons = pya.Region()
        for polygon in polygons.each():
            polygonreg = pya.Region(polygon)
            for ref in partmps_list:
                mps = ref.elem
                if ref.interacts_with(shapereg=polygonreg):
                    # Merge polygon
                    # print("\n", type(mps))
                    # print(mps.fullshape)
                    # print("+", polygonreg)
                    ref.add_polygon(polygon)
                    # print("=>", mps.fullshape)

                    merged_polygons += polygon # type: ignore
                    # Only merge it with one part, rest is handled ny merging
                    # part together.
                    break
        polygons -= merged_polygons

        # Join overlapping parts
        # Repeat until not mergers take place anymore; do at least one iteration
        merged = True
        while merged:
            merged = False
            # Merge all parts that overlap with another part and are not already merged
            for ref1, ref2 in combinations(partmps_list, 2):
                if not ref1.is_same_part_as(ref2):
                    if ref1.interacts_with(shapereg=ref2.region):
                        ref1.merge_with(ref2)
                        merged = True

        # Add remaining polygons
        shapes.extend(_ShapeRef(shape=_import_polygon(poly)) for poly in polygons.each())
        # Add converted parts
        filtered = filter(lambda ref: not ref.is_merged, partmps_list)
        shapes.extend(filtered)

        assert len(shapes) > 0, "Internal error"
        if len(shapes) == 1:
            result = shapes[0]
        else:
            result = _MultiShapeRef(shapes=tuple(shapes))
        return result

    def merge_sublayout(self, sl: _laylay._SubLayout) -> _SubLayoutReffed:
        if isinstance(sl, _laylay._MaskShapesSubLayout):
            return _MaskShapesSubLayoutRef(net=sl.net, ref=self(sl.shapes))
        else:
            # TODO: unit test with _InstanceSubLayout
            return _SubLayoutRef(sublayout=sl) # pragma: no cover

    def merge_layout(self, layout: _lay.LayoutT) -> _lay.LayoutT:
        # Create a tuple so we have merged all the sublayouts before we are going to deref
        # them
        slrefs = tuple(map(self.merge_sublayout, layout._sublayouts))
        sls = _laylay._SubLayouts(slref.deref() for slref in slrefs)
        return layout.fab.new_layout(sublayouts=sls, boundary = layout.boundary)

    def merge_shapes(self, shapes: Iterable[_geo._Shape]) -> Tuple[_ShapeReffed, ...]:
        return tuple(self(shape) for shape in shapes)

    def __enter__(self):
        return self

    def __exit__(self, *_):
        pass

    def __del__(self):
        # TODO: Check for all MultiPartShape._Part converted to new ones.
        # This is to avoid having shapes where some the the _Part objects point
        # to old MultiPartShape and some to a new one.
        pass


@overload
def merge(obj: _lay.LayoutT) -> _lay.LayoutT:
    ... # pragma: no cover
@overload
def merge(obj: _cell.Cell) -> None:
    ... # pragma: no cover
@overload
def merge(obj: _lbry.Library) -> None:
    ... # pragma: no cover
@overload
def merge(obj: _geo.MultiShape) -> _geo._Shape:
    ... # pragma: no cover
@overload
def merge(obj: _geo.MaskShape) -> _geo.MaskShape:
    ... # pragma: no cover
@overload
def merge(obj: _geo.MaskShapes) -> _geo.MaskShapes:
    ... # pragma: no cover
def merge(
    obj: Union[
        _lay.LayoutT, _cell.Cell, _lbry.Library, _geo.MultiShape,
        _geo.MaskShape, _geo.MaskShapes,
    ],
) -> Optional[Union[
    _lay.LayoutT, _geo._Shape, _geo.MaskShape, _geo.MaskShapes,
]]:
    """This function allows to use the KLayout engine to merge PDKMaster _Shape objects;
    either directly or as part of another PDKMaster object

    Althouogh it will be tried to merge shapes there is no guarantee given that merge-able
    shapes are effectively merged. Currently RepeatedShape objects are retained as is and
    not merged with other shapes.

    Arguments:
        obj: the PDKMaster object from which the shape(s) needs to be merged.
            obj can also be an iterable of Shape objects.

    Returns:
        Return same type of object; the object may have been simplified; e.g.
        merging a MultiShape with all shapes overlapping can result in a Polygon or a Rect.
        Currently RepeatedShape and _InstanceSubLayout will not be merge and returned
        as is.
        For object of type `_Cell` and `Library` the merging will be done on the provided
        object and no value is returned.
    """
    if isinstance(obj, _cell.Cell):
        for l in obj._layouts:
            l.layout = merge(l.layout)
    elif isinstance(obj, _lbry.Library):
        for cell in obj.cells:
            merge(cell)
    else:
        merged = _ShapeMerger()(obj)
        if isinstance(merged, _lay.LayoutT):
            return merged
        else:
            return merged.deref()



# SPDX-License-Identifier: AGPL-3.0-or-later OR GPL-2.0-or-later OR CERN-OHL-S-2.0+ OR Apache-2.0
from textwrap import dedent
from itertools import product
from xml.etree import ElementTree as ET
from typing import Any, Tuple, List, Dict, Set, Iterable, Optional, Union, cast, overload

from pdkmaster.technology import property_ as _prp

from pdkmaster import _util, dispatch as _dsp
from pdkmaster.typing import GDSLayerSpecDict
from pdkmaster.technology import (
    property_ as _prp, rule as _rle, wafer_ as _wfr, mask as _msk, edge as _edg,
    geometry as _geo, primitive as _prm, net as _net, technology_ as _tch,
)
from pdkmaster.design import layout as _lay, cell as _cell, library as _lbry
from pdkmaster.design.layout import layout_ as _laylay
from pdkmaster.io.spice import SpicePrimParamsT, SpicePrimsParamSpec

import pya


__all__= ["FileExporter", "export2db"]


class _MaskConverter(_dsp.MaskDispatcher):
    def __init__(self, *, tech: _tch.Technology) -> None:
        super().__init__()

        self.tech = tech

    def __call__(self, mask: _msk.MaskT) -> str:
        return super().__call__(mask)

    @staticmethod
    def __legalized_maskname(name: str) -> str:
        if name[0] in "0123456789":
            name = "_" + name
        return name.replace(".", "_").replace(":", "__")

    def DesignMask(self, mask: _msk.DesignMask) -> str:
        return self.__legalized_maskname(mask.name)

    def _MaskAlias(self, mask: _msk._MaskAlias) -> str:
        return self.__legalized_maskname(mask.name)

    # Handled on higher level
    # def _PartsWith(self, pw: msk._PartsWith):
    #     ...

    def Join(self, join: _msk.Join) -> str:
        return f"({'+'.join(self(m) for m in join.masks)})"

    def Intersect(self, intersect: _msk.Intersect) -> str:
        return f"({'&'.join(self(m) for m in intersect.masks)})"

    def _MaskRemove(self, mr: _msk._MaskRemove):
        return f"({self(mr.from_)}-{self(mr.what)})"

    def _Wafer(self, wafer: _wfr._Wafer, *args, **kwargs):
        # Size box to be sure it is big enough for substrate enclosure check
        bias = 0.0
        for act in self.tech.primitives.__iter_type__(_prm.WaferWire):
            for enc in (act.min_substrate_enclosure, act.min_substrate_enclosure_same_type):
                if enc is not None:
                    bias = max(bias, enc.max())

        return "extent" if bias < self.tech.grid/100 else f"extent.sized({bias:.6})"
_mask_conv: Any = None


class _EdgeConverter(_dsp.EdgeDispatcher):
    def __call__(self, edge: _edg.EdgeT) -> str:
        return super().__call__(edge)

    def MaskEdge(self, edge: _edg.MaskEdge) -> str:
        return f"{_mask_conv(edge.mask)}.edges"

    def _DualEdgeOperation(self, op: _edg._DualEdgeOperation) -> str:
        s_edge1 = self(op.edge1)
        if isinstance(op.edge2, _msk.MaskT):
            s_edge2 = _mask_conv(op.edge2)
        elif isinstance(op.edge2, _edg.EdgeT):
            s_edge2 = self(op.edge2)
        else: # pragma: no cover
            raise TypeError(f"[Internal error]Unexpected type for edge2 of {str(op)}")
        if op.operation == "interact_with":
            return f"{s_edge1}.interacting({s_edge2})"
        else: # pragma: no cover
            raise NotImplementedError(f"[Internal error]Operation: {op.operation}")

    def Join(self, join: _edg.Join) -> str:
        s_join = "+".join(
            _mask_conv(e) if isinstance(e, _msk.MaskT) else self(e)
            for e in join.edges
        )
        return f"({s_join})"

    def Intersect(self, intersect: _edg.Intersect) -> str:
        s_intersect = "&".join(
            _mask_conv(e) if isinstance(e, _msk.MaskT) else self(e)
            for e in intersect.edges
        )
        return f"({s_intersect})"
_edge_conv = _EdgeConverter()


def _str_designmask(mask: _msk.DesignMask, *, gds_layers: GDSLayerSpecDict):
    gds_layer = gds_layers[mask.name]
    if gds_layer is None:
        return ""
    if not isinstance(gds_layer, tuple):
        gds_layer = (gds_layer, 0)
    return f"{_mask_conv(mask)} = input{gds_layer}\n"


def _str_alias(mask: _msk._MaskAlias):
    return f"{_mask_conv(mask)} = {_mask_conv(mask.mask)}\n"


def _str_grid(mask: _msk.MaskT, grid: float):
    s_mask = _mask_conv(mask)
    return dedent(f"""
        {s_mask}.ongrid({grid}).output(
            "{s_mask} grid", "{s_mask} grid: {grid}µm"
        )
    """[1:])


class _RuleConverter(_dsp.RuleDispatcher):
    def __call__(self, rule: _rle.RuleT, *, allow_unimplented: bool=False) -> str:
        if allow_unimplented:
            try:
                s = super().__call__(rule)
            except NotImplementedError:
                s = "# Not supported\n"
        else:
            s = super().__call__(rule)

        return f"# {rule}\n{s}"

    def GreaterEqual(self, ge: _prp.Operators.GreaterEqual):
        left = ge.left
        right = ge.right
        if isinstance(left, _msk._MaskProperty):
            s_mask = _mask_conv(left.mask)
            prop = left.prop_name
            if prop in {"width", "space"}:
                return dedent(f"""
                    {s_mask}.{prop}({right}).output(
                        "{s_mask} {prop}", "{s_mask} minimum {prop}: {right}µm"
                    )
                """[1:])
            elif left.prop_name == "area":
                return dedent(f"""
                    {s_mask}.with_area(nil, {right}).output(
                        "{s_mask} area", "{s_mask} minimum area: {right}µm"
                    )
                """[1:])
            elif left.prop_name == "density":
                assert isinstance(right, float)
                return dedent(f"""
                    {s_mask}_mindens = polygon_layer
                    dens_check({s_mask}_mindens, {s_mask}, {right}, 1)
                    {s_mask}_mindens.output(
                        "{s_mask} density", "{s_mask} minimum density: {round(100*right)}%"
                    )
                """[1:])
        elif isinstance(left, _msk._DualMaskProperty):
            prop = left.prop_name
            if (
                (prop == "space")
                and (
                    isinstance(left.mask1, _msk._PartsWith)
                    or isinstance(left.mask2, _msk._PartsWith)
                )
            ):
                # Special code for handling width based spacing rules
                # Main objective if to support space tabels specified for primitives
                # Other application area are mainly untested
                if isinstance(left.mask1, _msk._PartsWith):
                    conds = left.mask1.condition
                    assert left.mask1.mask == left.mask2
                    s_mask = _mask_conv(left.mask1.mask)
                else:
                    assert isinstance(left.mask2, _msk._PartsWith)
                    conds = left.mask2.condition
                    assert left.mask2.mask == left.mask1
                    s_mask = _mask_conv(left.mask2.mask)

                width = None
                length = None
                for cond in conds:
                    assert isinstance(cond, _prp.Operators.GreaterEqual)
                    assert isinstance(cond.left, _msk._MaskProperty)
                    if cond.left.prop_name == "width":
                        width = cond.right
                    elif cond.left.prop_name == "length":
                        length = cond.right
                    else: # pragma: no cover
                        raise RuntimeError("Internal error")
                assert width is not None, "Internal error"
                if length is None:
                    return dedent(f"""
                        space4width_check({s_mask}, {width}, {right}).output(
                            "{s_mask} table spacing",
                            "Minimum {s_mask} spacing for {cond.right}µm width: {right}µm"
                        )
                    """[1:])
                else:
                    return dedent(f"""
                        space4width_check({s_mask}, {width}, {right}).output(
                            "[Warning]{s_mask} table spacing",
                            "Check minimum {s_mask} spacing for {width}µm width and {length}µm length: {right}µm"
                        )
                        space4widthlength_check({s_mask}, {width}, {length}, {right}).output(
                            "{s_mask} table spacing",
                            "Minimum {s_mask} spacing for {width}µm width and {length}µm length: {right}µm"
                        )
                    """[1:])

            s_mask1 = _mask_conv(left.mask1)
            s_mask2 = _mask_conv(left.mask2)
            if prop == "space":
                return dedent(f"""
                    {s_mask1}.separation({s_mask2}, {right}, square).output(
                        "{s_mask1}:{s_mask2} spacing",
                        "Minimum spacing between {s_mask1} and {s_mask2}: {right}µm"
                    )
                """[1:])
            elif prop == "extend_over":
                return dedent(f"""
                    extend_check({s_mask2}, {s_mask1}, {right}).output(
                        "{s_mask1}:{s_mask2} extension",
                        "Minimum extension of {s_mask1} of {s_mask2}: {right}µm"
                    )
                """[1:])
        elif isinstance(left, _msk._DualMaskEnclosureProperty):
            s_mask1 = _mask_conv(left.mask1)
            s_mask2 = _mask_conv(left.mask2)
            prop = left.prop_name
            if prop == "enclosed_by":
                # TODO: Proper typing for Property
                assert isinstance(right, _prp.Enclosure)
                if not right.is_assymetric:
                    return dedent(f"""
                        {s_mask2}.enclosing({s_mask1}, {right.first}).output(
                            "{s_mask2}:{s_mask1} enclosure",
                            "Minimum enclosure of {s_mask2} around {s_mask1}: {right.first}µm"
                        )
                    """[1:])
                else:
                    s_desc = (
                        f"Minimum enclosure of {s_mask2} around {s_mask1}: "
                        f"{right.min()}µm minimum, {right.max()}µm opposite"
                    )
                    return dedent(f"""
                        oppenc_check({s_mask1}, {s_mask2}, {right.min()}, {right.max()}).output(
                            "{s_mask2}:{s_mask1} asymmetric enclosure",
                            "{s_desc}"
                        )
                    """[1:])
        elif isinstance(left, _edg._EdgeProperty):
            s_edge = _edge_conv(left.edge)
            prop = left.prop_name
            if prop == "length":
                return dedent(f"""
                    {s_edge}.with_length(nil, {right}).output(
                        "{s_edge} length",
                        "Minimum length of {s_edge}: {right}µm"
                    )
                """[1:])
            elif prop == "space":
                return dedent(f"""
                    {s_edge}.space({right}).output(
                        "{s_edge} space",
                        "Minimum spacing between {s_edge}: {right}µm"
                    )
                """[1:])
        elif isinstance(left, _edg._DualEdgeProperty):
            s_edge1 = _edge_conv(left.edge1)
            s_edge2 = (
                _mask_conv(left.edge2)+".edges" if isinstance(left.edge2, _msk.MaskT)
                else _edge_conv(left.edge2)
            )
            prop = left.prop_name
            if prop == "enclosed_by":
                return dedent(f"""
                    {s_edge2}.enclosing({s_edge1}, {right}).output(
                        "{s_edge2}:{s_edge1} enclosure",
                        "Minimum enclosure of {s_edge2} around {s_edge1}: {right}µm"
                    )
                """[1:])

        raise NotImplementedError(f"GreateEqual rule '{ge}'") # pragma: no cover

    def Equal(self, eq: _prp.Operators.Equal) -> str:
        left = eq.left
        right = eq.right
        assert isinstance(right, left.value_type)
        if isinstance(left, _msk._MaskProperty):
            s_mask = _mask_conv(left.mask)
            prop = left.prop_name
            if prop == "width":
                return dedent(f"""
                    width_check({s_mask}, {right}).output(
                        "{s_mask} width", "{s_mask} width: {right}µm"
                    )
                """[1:])
            elif prop == "area":
                if round(right, 6) != 0.0:
                    raise ValueError("For area equal check value can only be 0.0")
                return f'{s_mask}.output("{s_mask} empty")\n'
        elif isinstance(left, _edg._EdgeProperty):
            s_edge = _edge_conv(left.edge)
            prop = left.prop_name
            if prop == "length":
                if round(right, 6) != 0.0:
                    raise ValueError("For length equal check value can only be 0.0")
                return f'{s_edge}.output("{s_edge} empty")\n'

        raise NotImplementedError(f"Equal rule '{eq}'") # pragma: no cover

    def SmallerEqual(self, se: _prp.Operators.SmallerEqual) -> str:
        left = se.left
        right = se.right
        if isinstance(left, _edg._EdgeProperty):
            s_edge = _edge_conv(left.edge)
            prop = left.prop_name
            if prop == "length":
                return dedent(f"""
                    {s_edge}.with_length({right}, nil).output(
                        "{s_edge} maximum length",
                        "Maximum length of {s_edge}: {right}µm"
                    )
                """[1:])

        raise NotImplementedError(f"SmallerEqual rule '{se}'") # pragma: no cover

    def _MaskAlias(self, alias: _msk._MaskAlias) -> str:
        return f"{_mask_conv(alias)} = {_mask_conv(alias.mask)}\n"

    def Connect(self, conn: _msk.Connect) -> str:
        return "".join(
            f"connect({_mask_conv(mask1)}, {_mask_conv(mask2)})\n"
            for mask1, mask2 in product(conn.mask1, conn.mask2)
        )
_rule_conv = _RuleConverter()


def _str_lvsresistor(res: _prm.Resistor, *, params: SpicePrimParamsT):
    s = f"# {res.name}\n"

    s_res = _mask_conv(res.mask)
    s_conn = _mask_conv(res.wire.conn_mask)

    sheetres = params["sheetres"]
    if sheetres is not None:
        s += dedent(f"""
            extract_devices(resistor("{res.name}", {sheetres}), {{
                "R" => {s_res}, "C" => {s_conn},
            }})
            same_device_classes("{res.name}", "RES")
        """[1:])

    return s


def _str_lvsdiode(tech: _tch.Technology, diode: _prm.Diode, spice_params: SpicePrimParamsT):
    s = f"# {diode.name}\n"

    is_n = any(impl.type_ == _prm.nImpl for impl in diode.implant)

    s_diode = _mask_conv(diode.mask)
    s_conn = _mask_conv(diode.wire.conn_mask)
    s_well = _mask_conv(
        diode.well.mask if diode.well is not None
        else tech.substrate_prim.mask
    )

    if is_n:
        s_p = s_well
        s_n = s_diode
        s_conn_port = "tC"
    else:
        s_n = s_well
        s_p = s_diode
        s_conn_port = "tA"

    s += dedent(f"""
        extract_devices(diode("{spice_params['model']}"), {{
            "P" => {s_p}, "N" => {s_n}, "{s_conn_port}" => {s_conn}
        }})
    """[1:])

    return s


def _str_lvsmosfet(tech: _tch.Technology, mosfet: _prm.MOSFET, spice_params: SpicePrimParamsT):
    s = f"# {mosfet.name}\n"

    s_sd = _mask_conv(mosfet.gate.active.conn_mask)
    s_gate = _mask_conv(mosfet.gate4mosfet.mask)
    s_bulk = _mask_conv(
        mosfet.well.mask if mosfet.well is not None
        else tech.substrate_prim.mask
    )
    s_poly = _mask_conv(mosfet.gate.poly.conn_mask)

    s += dedent(f"""
        extract_devices(mos4("{spice_params['model']}"), {{
            "SD" => {s_sd}, "G" => {s_gate}, "tG" => {s_poly}, "W" => {s_bulk},
        }})
    """[1:])

    return s


class FileExporter:
    def __init__(self, *,
        tech: _tch.Technology, export_name: Optional[str]=None,
        gds_layers: GDSLayerSpecDict, prims_spiceparams: SpicePrimsParamSpec,
    ):
        self.tech = tech
        self.export_name = tech.name if export_name is None else export_name
        self.gds_layers = gds_layers
        self.prims_spiceparams = prims_spiceparams

        global _mask_conv
        _mask_conv = _MaskConverter(tech=tech)

    def __call__(self, *, layerprops: Optional[str]=None):
        return {
            "drc": self._s_drc(),
            "ly_drc": self._ly_drc(),
            "extract": self._s_extract(),
            "ly_extract": self._ly_extract(),
            "lvs": self._s_lvs(),
            "ly_tech": self._ly_tech(layerprops=layerprops),
        }

    def _s_drc(self):
        s = dedent(f"""
            # Autogenerated file. Changes will be overwritten.

            source(ENV["SOURCE_FILE"])
            report("{self.export_name} DRC", ENV["REPORT_FILE"])

        """[1:])

        return s + self._s_drcrules()

    def _ly_drc(self):
        ly_drc = ET.Element("klayout-macro")
        ET.SubElement(ly_drc, "description")
        ET.SubElement(ly_drc, "version")
        ET.SubElement(ly_drc, "category").text = "drc"
        ET.SubElement(ly_drc, "prolog")
        ET.SubElement(ly_drc, "epilog")
        ET.SubElement(ly_drc, "doc")
        ET.SubElement(ly_drc, "autorun").text = "false"
        ET.SubElement(ly_drc, "autorun-early").text = "false"
        ET.SubElement(ly_drc, "shortcut")
        ET.SubElement(ly_drc, "show-in-menu").text = "true"
        ET.SubElement(ly_drc, "group-name").text = "drc_scripts"
        ET.SubElement(ly_drc, "menu-path").text = "tools_menu.drc.end"
        ET.SubElement(ly_drc, "interpreter").text = "dsl"
        ET.SubElement(ly_drc, "dsl-interpreter-name").text = "drc-dsl-xml"
        s = dedent(f"""
            # Autogenerated file. Changes will be overwritten.
            
            report("{self.export_name} DRC")

        """[1:]) + self._s_drcrules()
        ET.SubElement(ly_drc, "text").text = s
        
        return ly_drc

    def _s_drcrules(self):
        s = dedent(f"""
            def width_check(layer, w)
                small = layer.width(w).polygons
                big = layer.sized(-0.5*w).size(0.5*w)

                small | big
            end

            def space4width_check(layer, w, s)
                big = layer.sized(-0.5*w).size(0.5*w)
                big.edges.separation(layer.edges, s)
            end

            def space4widthlength_check(layer, w, l, s)
                big = layer.sized(-0.5*w).size(0.5*w)
                big.edges.separation(layer.edges, s).with_length(l + 1.dbu, nil)
            end

            def oppenc_check(inner, outer, min, max)
                toosmall = outer.enclosing(inner, min).second_edges

                smallenc = outer.enclosing(inner, max - 1.dbu, projection).second_edges
                # These edges may not touch each other
                touching = smallenc.width(1.dbu, angle_limit(100)).edges

                inner.interacting(toosmall + touching)
            end

            def extend_check(base, extend, e)
                extend.enclosing(base, e).first_edges.not_interacting(base)
            end

            def dens_check(output, input, min, max)
                tp = RBA::TilingProcessor::new

                tp.output("res", output.data)
                tp.input("input", input.data)
                tp.dbu = 1.dbu  # establish the real database unit
                tp.var("vmin", min)
                tp.var("vmax", max)

                tp.queue("_tile && (var d = to_f(input.area(_tile.bbox)) / to_f(_tile.bbox.area); (d < vmin || d > vmax) && _output(res, _tile.bbox))")
                tp.execute("Density check")
            end
        """[1:])

        s += "\n# Define layers\n"
        assert isinstance(self.tech.rules, _rle.Rules), "Internal error"
        dms = tuple(self.tech.rules.__iter_type__(_msk.DesignMask))
        s += "".join(_str_designmask(dm, gds_layers=self.gds_layers) for dm in dms)

        s += "\n# Grid check\n"
        gridrules = cast(Tuple[_prp.Operators.Equal, ...], tuple(filter(
                lambda rule: (
                    isinstance(rule, _prp.Operators.Equal)
                    and isinstance(rule.left, _msk._MaskProperty)
                    and (rule.left.prop_name == "grid")
                ),
                self.tech.rules,
            )
        ))
        gridspecs = {
            cast(_msk._MaskProperty, gridrule.left).mask: gridrule.right
            for gridrule in gridrules
        }
        globalgrid = gridspecs[_wfr.wafer]
        s += "".join(
            _str_grid(dm, cast(float, gridspecs.get(dm, globalgrid)))
            for dm in dms
        )

        s += "\n# Derived layers\n"
        aliases = tuple(self.tech.rules.__iter_type__(_msk._MaskAlias))
        s += "".join(_rule_conv(alias) for alias in aliases)

        s += "\n# Connectivity\n"
        conns = tuple(self.tech.rules.__iter_type__(_msk.Connect))
        s += "".join(_rule_conv(conn) for conn in conns)

        s += "\n# DRC rules\n" + "".join(
            _rule_conv(rule) for rule in filter(
                lambda rule: rule not in dms + gridrules + conns + aliases,
                self.tech.rules
            )
        )

        return s

    def _s_extract(self):
        s = dedent(f"""
            # Autogenerated file. Changes will be overwritten

            source(ENV["SOURCE_FILE"])
            target_netlist(ENV["SPICE_FILE"], write_spice(true, true))

        """[1:])
        s += self._s_extractrules()

        return s

    def _ly_extract(self):
        ly_extract = ET.Element("klayout-macro")
        ET.SubElement(ly_extract, "description")
        ET.SubElement(ly_extract, "version")
        ET.SubElement(ly_extract, "category").text = "lvs"
        ET.SubElement(ly_extract, "prolog")
        ET.SubElement(ly_extract, "epilog")
        ET.SubElement(ly_extract, "doc")
        ET.SubElement(ly_extract, "autorun").text = "false"
        ET.SubElement(ly_extract, "autorun-early").text = "false"
        ET.SubElement(ly_extract, "shortcut")
        ET.SubElement(ly_extract, "show-in-menu").text = "true"
        ET.SubElement(ly_extract, "group-name").text = "lvs_scripts"
        ET.SubElement(ly_extract, "menu-path").text = "tools_menu.lvs.end"
        ET.SubElement(ly_extract, "interpreter").text = "dsl"
        ET.SubElement(ly_extract, "dsl-interpreter-name").text = "lvs-dsl-xml"
        s = dedent(f"""
            # Autogenerated file. Changes will be overwritten

            report_netlist

        """[1:])
        s += self._s_extractrules()
        ET.SubElement(ly_extract, "text").text = s
        
        return ly_extract

    def _s_lvs(self):
        s = dedent(f"""
            # Autogenerated file. Changes will be overwritten

            source(ENV["SOURCE_FILE"])
            schematic(ENV["SPICE_FILE"])
            report_lvs(ENV["REPORT_FILE"])

        """[1:])
        s += self._s_extractrules() + dedent(f"""
            align
            ok = compare
            if ok then
                print("LVS OK\\n")
            else
                abort "LVS Failed!"
            end
        """)

        return s

    def _s_extractrules(self):
        # TODO: bug report for failing LVS on hierarchical LVS and diodes
        s = "flat\n\n# Define layers\n"
        assert isinstance(self.tech.rules, _rle.Rules), "Internal error"
        dms = tuple(self.tech.rules.__iter_type__(_msk.DesignMask))
        s += "".join(_str_designmask(dm, gds_layers=self.gds_layers) for dm in dms)
        aliases = tuple(self.tech.rules.__iter_type__(_msk._MaskAlias))
        s += "".join(_str_alias(alias) for alias in aliases)

        s += "\n# Connectivity\n"
        conns = tuple(self.tech.rules.__iter_type__(_msk.Connect))
        s += "".join(_rule_conv(conn) for conn in conns)

        s += "\n# Resistors\n"
        resistors = tuple(self.tech.primitives.__iter_type__(_prm.Resistor))
        s += "".join(
            _str_lvsresistor(res, params=self.prims_spiceparams[res])
            for res in resistors
        )

        s += "\n# Diodes\n"
        diodes = tuple(self.tech.primitives.__iter_type__(_prm.Diode))
        s += "".join(
            _str_lvsdiode(self.tech, diode, self.prims_spiceparams[diode])
            for diode in diodes
        )

        s += "\n# Transistors\n"
        mosfets = tuple(self.tech.primitives.__iter_type__(_prm.MOSFET))
        s += "".join(
            _str_lvsmosfet(self.tech, mosfet, self.prims_spiceparams[mosfet])
            for mosfet in mosfets
        )

        s += "\nnetlist\n"

        return s

    def _ly_tech(self, *, layerprops: Optional[str]):
        lyt = ET.Element("technology")
        ET.SubElement(lyt, "name").text = self.export_name
        ET.SubElement(lyt, "description").text = (
            f"KLayout generated from {self.tech.name} PDKMaster technology"
        )
        ET.SubElement(lyt, "group")
        ET.SubElement(lyt, "dbu").text = f"{self.tech.dbu}"
        se = ET.SubElement(lyt, "layer-properties_file")
        if layerprops is not None:
            se.text = layerprops
        ET.SubElement(lyt, "add-other-layers").text = "true"
        ropts = ET.SubElement(lyt, "reader-options")
        roptscom = ET.SubElement(ropts, "common")
        ET.SubElement(roptscom, "create-other-layers").text = "true"
        def s_gds_layer(m: _msk.DesignMask):
            try:
                l = self.gds_layers[m.name]
            except KeyError: # pragma: no cover
                raise ValueError(
                    f"No gds_layer provided for mask '{m.name}'"
                )
            if l is None:
                return None
            elif isinstance(l, int):
                return f"{l}/0"
            else:
                assert isinstance(l, tuple)
                return f"{l[0]}/{l[1]}"
        s_map = ";".join(
            f"'{s_gds_layer(mask)} : {mask.name}'"
            for mask in self.tech.designmasks if s_gds_layer(mask) is not None
        )
        ET.SubElement(roptscom, "layer-map").text = f"layer_map({s_map})"
        ET.SubElement(lyt, "writer-options")
        ET.SubElement(lyt, "connectivity")

        return lyt


class _ShapeExporter(_dsp.ShapeDispatcher):
    """Converts a _geo,_Shape object to KLayout database object"""
    def __init__(self, *, export_fullshape: bool):
        self._mps_exported: Optional[Set[_geo.MultiPartShape]]
        self._mps_exported = set() if export_fullshape else None

    def _pointsshapes(self, shape: _geo._Shape) -> Iterable[Any]:
        # Helper to convert the individual pointsshapes
        for pointshape in shape.pointsshapes:
            conv = self(pointshape)
            assert not _util.is_iterable(conv), "Internal error: unsupported"
            yield conv

    def _Shape(self, shape: _geo._Shape): # pragma: no cover
        raise ValueError(f"Unsupported object of type {shape.__class__.__name__}")

    def Point(self, point: _geo.Point) -> pya.DPoint:
        return pya.DPoint(point.x, point.y)

    def Line(self, line: _geo.Line) -> pya.DPath:
        # We represent a Line by a path with zero width
        points = (self.Point(line.point1), self.Point(line.point2))
        return pya.DPath(points, 0.0)

    def Polygon(self, polygon: _geo.Polygon, **_) -> pya.DSimplePolygon:
        # In PDKMaster polygon needs last point to be same as first point;
        # in klayout this is not the case.
        points = tuple(self.Point(point) for point in tuple(polygon.points)[:-1])
        return pya.DSimplePolygon(points)

    def Rect(self, rect: _geo.Rect) -> pya.DBox:
        return pya.DBox(rect.left, rect.bottom, rect.right, rect.top)

    def RectRing(self, rs: _geo.RectRing) -> Iterable[Any]:
        # TODO: Can repetition information be retained in KLayout ?
        return self._pointsshapes(rs)

    def Label(self, label: _geo.Label) -> pya.DText:
        return pya.DText(label.text, label.origin.x, label.origin.y)

    def MultiPartShape(self, mps: _geo.MultiPartShape):
        if self._mps_exported is None:
            return self(mps.fullshape)
        else:
            if mps in self._mps_exported:
                return None
            else:
                self._mps_exported.add(mps)
                return self(mps.fullshape)

    def MultiPartShape__Part(self, part: _geo.MultiPartShape._Part):
        if self._mps_exported is None:
            return self(part.partshape)
        else:
            return self(part.multipartshape)

    def MultiShape(self, ms: _geo.MultiShape) -> Iterable[Any]:
        for shape in ms.shapes:
            conv = self(shape)
            if _util.is_iterable(conv):
                yield from conv
            else:
                yield conv

    def RepeatedShape(self, rs: _geo.RepeatedShape) -> Iterable[Any]:
        # TODO: Can repetition information be retained in KLayout ?
        return self._pointsshapes(rs)

    # TODO: Does KLayout allow more efficient array representation
    # ArrayShape -> RepeatedShape


class _MaskLayerDict(Dict[_msk.DesignMask, int]):
    def __init__(self, *, layout: pya.Layout, gds_layers: GDSLayerSpecDict):
        self._layout = layout
        self._gds_layers = gds_layers

    def __getitem__(self, mask: _msk.DesignMask) -> int:
        if mask not in self:
            layer = self._gds_layers[mask.name]
            if isinstance(layer, tuple):
                layer, datatype = layer
            else:
                datatype = 0
            assert layer is not None
            self[mask] = self._layout.layer(layer, datatype, mask.name)
        return super().__getitem__(mask)


_rotation_to_rot_mirr: Dict[_geo.Rotation, Tuple[int, bool]] = {
    _geo.Rotation.No: (0, False),
    _geo.Rotation.R90: (1, False),
    _geo.Rotation.R180: (2, False),
    _geo.Rotation.R270: (3, False),
    _geo.Rotation.MX: (0, True),
    _geo.Rotation.MX90:(1, True),
    _geo.Rotation.MY: (2, True),
    _geo.Rotation.MY90: (3, True),
}


class _LayoutExporter:
    def __init__(self):
        self._clear()

    def _clear(self):
        self.layout = None
        self.layerdict = None
        self.textlayerdict = None
        self.cell_lookup: Dict[str, Tuple[Optional[_cell.Cell], pya.Cell]] = {}
        self.cells_todo: Set[_cell.Cell] = set()
        self.cells_done: Set[_cell.Cell] = set()
        self.cell = None
        self.shapeexporter = None
        self.pinmasks: Optional[Tuple[_msk.MaskT, ...]] = None

    def __call__(self, *,
        obj: Union[_geo.MaskShape, _geo.MaskShapes, _lay.LayoutT, _lbry.Library],
        pya_layout: Optional[pya.Layout]=None,
        gds_layers: GDSLayerSpecDict,
        textgds_layers: Optional[GDSLayerSpecDict]=None,
        cell_name: Optional[str], pya_cell: Optional[pya.Cell],
        merge: bool, add_pin_label: bool, dbu: float=0.001,
    ) -> pya.Layout:
        if pya_layout is None:
            pya_layout = pya.Layout() if pya_cell is None else pya_cell.layout()
        self.layout = pya_layout
        pya_layout.dbu = dbu
        self.layerdict = _MaskLayerDict(layout=pya_layout, gds_layers=gds_layers)
        if textgds_layers is not None:
            self.textlayerdict = _MaskLayerDict(layout=pya_layout, gds_layers=textgds_layers)
        self.shapeexporter = _ShapeExporter(export_fullshape=True)

        if isinstance(obj, _lbry.Library):
            # A cell will be created in add() function below for each cell of the library
            if cell_name is not None:
                raise TypeError("cell_name may not be specified when exporting a library")
            if pya_cell is not None:
                raise TypeError("pya_cell may not be specified when exporting a library")
        else:
            if cell_name is None:
                cell_name = "anon"
            if pya_cell is None:
                pya_cell = self._create_layout(cell_name)
            else:
                self.cell_lookup[cell_name] = (None, pya_cell)
            self.cell = pya_cell

        # Define local function to allow recursive calling
        self._add(obj, add_pin_label=add_pin_label, net=None)
        while len(self.cells_todo) > 0:
            cell = self.cells_todo.pop()
            self.cells_done.add(cell)
            self.cell = self.cell_lookup[cell.name][1]
            # Don't reuse ShapeExporter between cells, otherwise MultiPartShapes
            # can be wrongly marked as written when they are not.
            self.shapeexporter = _ShapeExporter(export_fullshape=True)
            self._add(cell.layout, add_pin_label=add_pin_label, net=None)

        if merge:
            for cell in pya_layout.each_cell():
                for layer_idx in self.layerdict.values():
                    # https://www.klayout.de/forum/discussion/697/merge-all-shapes-of-a-certain-layer-of-a-cell
                    old_shapes = cell.shapes(layer_idx)
                    region = pya.Region(old_shapes)
                    region.merge()

                    new_shapes = pya.Shapes()
                    new_shapes.insert(region)
                    # Copy texts over to avoid they are lost
                    for shape in old_shapes.each():
                        if shape.is_text():
                            new_shapes.insert(shape)

                    cell.shapes(layer_idx).assign(new_shapes)

        self._clear()
        return pya_layout

    def _create_layout(self, name: str):
        assert self.layout is not None
        assert len(self.cell_lookup) == 0
        cell = self.layout.create_cell(name)
        self.cell_lookup[name] = (None, cell)

        return cell

    def _register_cell(self, cell: _cell.Cell):
        assert self.layout is not None
        assert self.cell_lookup is not None

        name = cell.name
        if name in self.cell_lookup:
            # Check if no two cells from different libraries with the same name are used
            if self.cell_lookup[name][0] != cell: # pragma: no cover
                raise ValueError(
                    "Export of hierarchy with two cells from different libraries with same name not supported"
                )
        else:
            self.cell_lookup[name] = (cell, self.layout.create_cell(name))
            if cell not in self.cells_done:
                self.cells_todo.add(cell)
        return self.cell_lookup[name][1]

    def _add(self,
        o: Union[_geo.MaskShape, _geo.MaskShapes, _lay.LayoutT, _cell.Cell, _lbry.Library],
        add_pin_label: bool, net: Optional[_net.NetT],
    ) -> None:
        assert self.shapeexporter is not None
        if isinstance(o, _geo.MaskShape):
            assert (self.cell is not None) and (self.layerdict is not None)
            layer = self.layerdict[o.mask]
            shapes = self.cell.shapes(layer)
            exp = self.shapeexporter(o.shape)
            if _util.is_iterable(exp):
                for s in exp:
                    # s is None if MultiPartShape is already added
                    if s is not None:
                        shapes.insert(s)
            else:
                # exp is None if MultiPartShape has already been converted
                if exp is not None:
                    shapes.insert(exp)
            if add_pin_label:
                assert self.pinmasks is not None
                if o.mask in self.pinmasks:
                    if self.textlayerdict is None:
                        text_shapes = shapes
                    else:
                        try:
                            layer = self.textlayerdict[o.mask]
                        except:
                            text_shapes = shapes
                        else:
                            text_shapes = self.cell.shapes(layer)
                    assert net is not None
                    for ps in o.shape.pointsshapes:
                        if isinstance(ps, _geo.Rect):
                            point = ps.center
                        else:
                            point = _util.get_first_of(ps.points)
                        text_shapes.insert(pya.DText(net.name, point.x, point.y))
        elif isinstance(o, _geo.MaskShapes):
            for ms in o:
                self._add(ms, add_pin_label=add_pin_label, net=net)
        elif isinstance(o, _lay.LayoutT):
            prims = o.fab.tech.primitives
            pinmasks: List[_msk.DesignMask] = []
            for prim in prims:
                try:
                    pin: _prm.Marker = prim.pin # type: ignore
                except AttributeError:
                    pass
                else:
                    pinmasks.append(pin.mask)
            self.pinmasks = tuple(pinmasks)
            for sl in o._sublayouts:
                if isinstance(sl, _laylay._MaskShapesSubLayout):
                    self._add(sl.shapes, add_pin_label=add_pin_label, net=sl.net)
                elif isinstance(sl, _laylay._InstanceSubLayout):
                    assert self.cell is not None
                    instcell = self._register_cell(sl.inst.cell)
                    rot, mirr = _rotation_to_rot_mirr[sl.rotation]
                    dtrans = pya.DTrans(rot, mirr, sl.origin.x, sl.origin.y)
                    self.cell.insert(pya.DCellInstArray(instcell.cell_index(), dtrans))
                else: # pragma: no cover
                    raise RuntimeError(
                        "Internal error: unsupported",
                    )
        elif isinstance(o, _cell.Cell):
            assert net is None
            # Register cell to be exported
            self._register_cell(o)
            # Make use of self.cell after this an error
            self.cell = None
        elif isinstance(o, _lbry.Library):
            # Initialize list of cells to tape-out, each of the cells will be
            # added in a loop below this function.
            assert net is None
            self.cells_todo.update(o.cells)
            # Register all cells to be exported
            for lbrycell in o.cells:
                self._register_cell(lbrycell)
            # Make use of self.cell after this an error
            self.cell = None
        else:
            raise TypeError(f"No support for exporting a '{type(o)}' object")


@overload
def export2db(
    obj: _geo._Shape, *,
    pya_layout: Optional[pya.Layout]=None,
    export_fullshape: Optional[bool]=None,
    add_pin_label: bool=False,
    gds_layers: None=None,
    textgds_layers: None=None,
    cell_name: Optional[str]=None, pya_cell: Optional[pya.Cell]=None,
    merge: bool=False,
) -> Any:
    ... # pragma: no cover
@overload
def export2db(
    obj: Union[_geo.MaskShape, _geo.MaskShapes, _lay.LayoutT, _lbry.Library], *,
    pya_layout: Optional[pya.Layout]=None,
    export_fullshape: None=None,
    add_pin_label: bool=False,
    gds_layers: GDSLayerSpecDict,
    textgds_layers: Optional[GDSLayerSpecDict]=None,
    cell_name: Optional[str]=None, pya_cell: Optional[pya.Cell]=None,
    merge: bool=False,
) -> pya.Layout:
    ... # pragma: no cover
def export2db(
    obj: Union[_geo._Shape, _geo.MaskShape, _geo.MaskShapes, _lay.LayoutT, _lbry.Library], *,
    pya_layout: Optional[pya.Layout]=None,
    export_fullshape: Optional[bool]=None, add_pin_label: bool=False,
    gds_layers: Optional[GDSLayerSpecDict]=None,
    textgds_layers: Optional[GDSLayerSpecDict]=None,
    cell_name: Optional[str]=None, pya_cell: Optional[pya.Cell]=None,
    merge: bool=False,
):
    """This function allows to export PDKMaster geometry/layout objects
    to a klayout Layout object

    Arguments:
        obj: This is the object to export. There are two different call types.
            1) a _Shape geometry object without mask provided
            2) a MaskShape geometry or an object that contains maskshapes.
        pya_layout: Optional existing KLayout layout to use for export.
            If not specified a fresh layout will be created.
        export_fullshape: only to be specified when obj is a _Shape object.
            If `True` the full MultiPartShape shape will be exported when a
            MultiPartShape._Part object is met. It defaults to `False`
        add_pin_label: If `True` a label will exported on top of layers that are pin
            layers of one of the technology's MetalWire primitives.
        gds_layers: Has to be specified when `obj` is a MaskShape or a collection of them
            and should not be provided otherwise.
            It contains the lookup table to get the corresponding KLayout layer information
            for PDKMaster _DesignMask objects.
        textgds_layers: Optional for when `obj` is a MaskShape.
            It's an extra lookup table for the layer to generate text for pins in.
            If no value is found or textgds_layers is not specified the layer from
            gds_layers will be used.
        cell_name: Only to be provided when `obj` is a MaskShape or a collection of them but
            not a Library.
            If specified the name of the cell in which the MaskShape(s) will be exported.
            By default 'anon' will be used.
        pya_cell: Optional KLayout cell to export shapes into.
            May not be specified when exporting a library.
            If not specified otherwise, a cell with name given by cell_name will be created.
        merge: Wether to merge the exported shapes or not.

    Returns:
        An equivalent KLayout database object if obj is a _Shape geometry or a KLayout
        Layout object when obj is a MaskShape or a collection of MaskShapes. If obj is
        a Library a Cell will be added for each _Cell in the Library. If there are
        cell instances in the PDKMaster _Layout object these cells will also be exported
        to the output KLayout Layout object as a cell even if it is from another PDKMaster
        Library. An exception will be generated when two instances need to be exported to
        the same cell name but in a different Library.
    """
    # TODO: Provide facility to also export netlist information

    if isinstance(obj, _geo._Shape):
        assert (gds_layers is None) and (cell_name is None) and (not add_pin_label)
        if export_fullshape is None:
            export_fullshape = False
        _exporter = _ShapeExporter(export_fullshape=export_fullshape)
        return _exporter(obj)
    else:
        assert (export_fullshape is None) and (gds_layers is not None)
        return _LayoutExporter()(
            obj=obj, pya_layout=pya_layout, gds_layers=gds_layers, textgds_layers=textgds_layers,
            cell_name=cell_name, pya_cell=pya_cell, merge=merge, add_pin_label=add_pin_label,
        )



# SPDX-License-Identifier: AGPL-3.0-or-later OR GPL-2.0-or-later OR CERN-OHL-S-2.0+ OR Apache-2.0
from itertools import product
from typing import List, Tuple, Dict, Optional, Any, cast
import pya

from pdkmaster import typing as _typ
from pdkmaster.technology import geometry as _geo, property_ as _prp, primitive as _prm
from pdkmaster.design import circuit as _ckt, layout as _lay
from pdkmaster.dispatch import PrimitiveDispatcher

from .export import export2db


__all__ = ["PCellLibrary"]


# Some notes about use of pya.PCellDeclarationHelper
# Internal klayout code does not seem to like it when assigning value to field with names
# other than a parameter name. It seems to be OK if it happens in __init__() so for the
# moment we will use a vals dict to store values.
# When a parameter is defined the type of the field (e.g. obj.param) varies between
# different methods. In such case the type of the parameter may need to be defined as Any.

# Currently the types in PCellDeclarationHelper are not annotated
_TypeDouble: int = pya.PCellDeclarationHelper.TypeDouble # type: ignore
_TypeInt: int = pya.PCellDeclarationHelper.TypeInt # type: ignore
_TypeBoolean: int = pya.PCellDeclarationHelper.TypeBoolean # type: ignore


class _MOSFET(pya.PCellDeclarationHelper):
    def __init__(self, *,
        layoutfab: _lay.LayoutFactory, gds_layers: _typ.GDSLayerSpecDict,
        mosfet: _prm.MOSFET,
    ):
        self._layoutfab = layoutfab
        self._gds_layers = gds_layers
        self._mosfet = mosfet
        super().__init__()

        self.param("_w",  _TypeDouble, "Width", unit="µm", default=mosfet.computed.min_w)
        self._w: float
        self.param("_l",  _TypeDouble, "Length", unit="µm", default=mosfet.computed.min_l)
        self._l: float

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return f"{self._mosfet.name}(w={self._w:1.3f}, l={self._l:1.3f})"

    def coerce_parameters_impl(self):
        mosfet = self._mosfet
        l = self._l
        w = self._w

        min_l = mosfet.computed.min_l
        if l < (min_l - _geo.epsilon):
            self._l = min_l

        min_w = mosfet.computed.min_w
        if w < (min_w - _geo.epsilon):
            self._w = min_w

    def produce_impl(self):
        mosfet = self._mosfet

        lay = self._layoutfab.layout_primitive(mosfet, l=self._l, w=self._w)

        export2db(lay,
            gds_layers=self._gds_layers,
            pya_cell=self.cell,
            merge=True,
        )


class _MOSFETFingers(pya.PCellDeclarationHelper):
    def __init__(self, *,
        layoutfab: _lay.LayoutFactory, gds_layers: _typ.GDSLayerSpecDict,
        mosfet: _prm.MOSFET,
    ):
        self._layoutfab = layoutfab
        self._gds_layers = gds_layers
        self._mosfet = mosfet

        diff = mosfet.gate.active
        cont: Optional[_prm.Via] = None
        for via in layoutfab.tech.primitives.__iter_type__(_prm.Via):
            if diff in via.bottom:
                cont = via
                break
        assert cont is not None
        self._cont = cont

        self._min_l = mosfet.computed.min_l
        idx = cont.bottom.index(diff)
        enc = cont.min_bottom_enclosure[idx].max()
        self._min_w = max(
            mosfet.computed.min_w,
            cont.width + 2*enc,
        )

        super().__init__()

        self.param("_w",  _TypeDouble, "Width", unit="µm", default=mosfet.computed.min_w)
        self._w: float
        self.param("_l",  _TypeDouble, "Length", unit="µm", default=mosfet.computed.min_l)
        self._l: float

        self.param("_fingers", _TypeInt, "Finger", default=2)
        self._fingers: int
        self.param("_left", _TypeBoolean, "Left cont.", default=True)
        self._left: bool
        self.param("_in", _TypeBoolean, "Inner cont.", default=True)
        self._in: bool
        self.param("_right", _TypeBoolean, "Right cont.", default=True)
        self._right: bool

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return f"{self._mosfet.name}(w={self._w:1.3f}, l={self._l:1.3f}, m={self._fingers})"

    def coerce_parameters_impl(self):
        mosfet = self._mosfet
        l = self._l
        min_l = self._min_l
        w = self._w
        min_w = self._min_w

        if l < (min_l - _geo.epsilon):
            self._l = min_l
        if w < (min_w - _geo.epsilon):
            self._w = min_w

    def produce_impl(self):
        layoutfab = self._layoutfab
        mosfet = self._mosfet
        cont = self._cont

        l = self._l
        w = self._w
        fingers = self._fingers
        leftch = self._left
        inch = self._in
        rightch = self._right

        cktfab = _ckt.CircuitFactory(tech=layoutfab.tech)

        ckt = cktfab.new_circuit(name="trans")
        layouter = layoutfab.new_circuitlayouter(circuit=ckt, boundary=None)

        def get_spec(i: int):
            inst = ckt.instantiate(mosfet, name=f"mos{i}", l=l, w=w)
            vl = cont if (((i == 0) and leftch) or ((i > 0) and inch)) else None
            vr = cont if (
                ((i < (fingers - 1)) and inch)
                or ((i == (fingers - 1)) and rightch)
            ) else None
            return _lay.MOSFETInstSpec(inst=inst, contact_left=vl, contact_right=vr)
        specs = tuple(get_spec(i) for i in range(fingers))

        ckt.new_net(name="g", external=False, childports=(
            *(spec.inst.ports["gate"] for spec in specs),
        ))
        ckt.new_net(name="sd", external=False, childports=(
            *(spec.inst.ports["sourcedrain1"] for spec in specs),
            *(spec.inst.ports["sourcedrain2"] for spec in specs),
        ))
        ckt.new_net(name="b", external=False, childports=(
            *(spec.inst.ports["bulk"] for spec in specs),
        ))

        layout = layouter.transistors_layout(trans_specs=specs)

        export2db(layout,
            gds_layers=self._gds_layers,
            pya_cell=self.cell,
            merge=True,
        )



class _ViaArray(pya.PCellDeclarationHelper):
    def __init__(self, *,
        layoutfab: _lay.LayoutFactory, gds_layers: _typ.GDSLayerSpecDict,
        via: _prm.Via,
    ):
        self._layoutfab = layoutfab
        self._gds_layers = gds_layers
        self._via = via
        self._vals = vals = {}
        super().__init__()

        self.param("_enct", _TypeInt, "Sizing", default=0, choices=(
            ("Minimal", 0), ("Enclosure", 1), ("Absolute", 2),
        ))
        self._enct: Any

        self.param("_rows", _TypeInt, "Rows", default = 2)
        self._rows: Any
        self.param("_cols", _TypeInt, "Columns", default = 1)
        self._cols: Any

        enc_choices = (("Wide", 0), ("Tall", 1))

        vals["n_diff"] = None
        choices = tuple((b.name, n) for n, b in enumerate(via.bottom))
        self.param("_b", _TypeInt, "Bottom", default=0, choices=choices, hidden=(len(via.bottom) == 1))
        self._b: Any

        enc = via.min_bottom_enclosure[0]
        hidden = not enc.is_assymetric
        d = 0 if enc.first > enc.second else 1
        self.param(
            "_benc", _TypeInt, "Bottom enclosure", default=d, choices=enc_choices, hidden=hidden,
        )
        self.param(
            "_benc_h", _TypeDouble, "Bot. hor. enclosure", unit="µm",
            default=enc.first, hidden=True,
        )
        self.param(
            "_benc_v", _TypeDouble, "Bot. ver. enclosure", unit="µm",
            default=enc.second, hidden=True,
        )
        self._benc: Any
        self._benc_h: Any
        self._benc_v: Any
        vals["benc"] = enc

        diffs = tuple(p for p in via.bottom if isinstance(p, _prm.WaferWire))
        if diffs:
            assert len(diffs) == 1
            diff = diffs[0]
            vals["n_diff"] = via.bottom.index(diff)
            choices = tuple(
                (im.name, n)
                for n, im in enumerate(diff.implant)
                if im.type_ in (_prm.nImpl, _prm.pImpl)
            )
            self.param(
                "_im", _TypeInt, "Bottom implant", default=0, choices=choices,
                hidden=(vals["n_diff"] != 0),
            )

            enc = diff.min_implant_enclosure[0]
            hidden = not enc.is_assymetric
            d = 0 if enc.first > enc.second else 1
            self.param(
                "_imenc", _TypeInt, "Impl.-diff. enclosure", unit="µm",
                default=d, choices=enc_choices, hidden=hidden,
            )
            self.param(
                "_imenc_h", _TypeDouble, "Impl. hor. enclosure", unit="µm",
                default=enc.first, hidden=True,
            )
            self.param(
                "_imenc_v", _TypeDouble, "Impl. ver. enclosure", unit="µm",
                default=enc.second, hidden=True,
            )
            vals["imenc"] = enc
        else:
            self.param(
                "_im", _TypeInt, "Bottom implant", default=0, hidden=True,
            )
            self.param(
                "_imenc", _TypeInt, "Impl.-diff. enclosure", unit="µm",
                default=0, choices=enc_choices, hidden=True,
            )
            self.param(
                "_imenc_h", _TypeDouble, "Bot. hor. enclosure", unit="µm",
                default=0.0, hidden=True,
            )
            self.param(
                "_imenc_v", _TypeDouble, "Bot. ver. enclosure", unit="µm",
                default=0.0, hidden=True,
            )
        self._im: Any
        self._imenc: Any
        self._imenc_h: Any
        self._imenc_v: Any

        choices = tuple((b.name, n) for n, b in enumerate(via.top))
        self.param("_t", _TypeInt, "Top", default=0, choices=choices, hidden=(len(via.top) == 1))
        self._t: Any

        enc = via.min_top_enclosure[0]
        hidden = not enc.is_assymetric
        d = 0 if enc.first > enc.second else 1
        self.param(
            "_tenc", _TypeInt, "Top enclosure", unit="µm",
            default=d, choices=enc_choices, hidden=hidden,
        )
        self.param(
            "_tenc_h", _TypeDouble, "Top hor. enclosure", unit="µm",
            default=enc.first, hidden=True,
        )
        self.param(
            "_tenc_v", _TypeDouble, "Top ver. enclosure", unit="µm",
            default=enc.second, hidden=True,
        )
        self._tenc: Any
        self._tenc_h: Any
        self._tenc_v: Any
        vals["tenc"] = enc

        benc = via.min_bottom_enclosure[0].max()
        tenc = via.min_top_enclosure[0].max()
        w = via.width + max(2*benc, 2*tenc)
        h = 2*via.width + via.min_space + max(2*benc, 2*tenc)
        self.param(
            "_padw", _TypeDouble, "Pad width", unit="µm", default=w, hidden=True,
        )
        self.param(
            "_padh", _TypeDouble, "Pad height", unit="µm", default=h, hidden=True,
        )
        self._padw: Any
        self._padh: Any

        self.param("_mins", _TypeBoolean, "Min. Space", default=True)
        self._mins: Any
        self.param(
            "_space", _TypeDouble, "Space", unit="µm", default=via.min_space, hidden=True,
        )
        self._space: Any

    def display_text_impl(self):
        via = self._via
        bottom = via.bottom[self._b]
        top = via.top[self._t]
        enct: int = self._enct

        if enct in (0, 1):
            return f"{bottom.name}-{via.name}-{top.name} ({self._rows}X{self._cols})"
        else:
            assert enct == 2
            padw = round(self._padw, 6)
            padh = round(self._padh, 6)
            return f"{bottom.name}-{via.name}-{top.name} ({padw}X{padh})"

    def coerce_parameters_impl(self):
        via = self._via
        enct: int = self._enct
        b: int = self._b
        t: int = self._t
        space = self._space

        if (enct == 2):
            encm = max(via.min_bottom_enclosure[b].max(), via.min_top_enclosure[t].max())
            d = via.width + 2*encm

            padw: float = self._padw
            if d > (padw + _geo.epsilon):
                self._padw = d

            padh: float = self._padh
            d = via.width + 2*encm
            if d > (padh + _geo.epsilon):
                self._padh = d

        if (space < (via.min_space - _geo.epsilon)):
            self._space = via.min_space

    # did not find way yet to test this code from unit test; done interactively in KLayout for now
    def callback_impl(self, name) -> None: # pragma: nocover
        via = self._via
        vals = self._vals
        n_diff = vals["n_diff"]

        b: int = self._b.value
        t: int = self._t.value
        enct = self._enct.value

        if name in ("", "_enct"):
            self._rows.visible = (enct != 2)
            self._cols.visible = (enct != 2)

            enc = via.min_bottom_enclosure[b]
            self._benc.visible = (enct == 0) and enc.is_assymetric
            self._benc_h.visible = (enct == 1)
            self._benc_v.visible = (enct == 1)

            if n_diff is not None:
                im: int = self._im.value

                enc = cast(_prm.WaferWire, via.bottom[n_diff]).min_implant_enclosure[im]
                self._imenc.visible = (n_diff == b) and enc.is_assymetric and (enct == 0)
                self._imenc_h.visible = (n_diff == b) and (enct == 1)
                self._imenc_v.visible = (n_diff == b) and (enct == 1)

            enc = via.min_top_enclosure[t]
            self._tenc.visible = (enct == 0) and enc.is_assymetric
            self._tenc_h.visible = (enct == 1)
            self._tenc_v.visible = (enct == 1)

            self._padw.visible = (enct == 2)
            self._padh.visible = (enct == 2)

        if name in ("", "_b"):
            enc = via.min_bottom_enclosure[b]
            self._benc.visible = (enct == 0) and enc.is_assymetric
            self._benc_h.visible = (enct == 1)
            self._benc_v.visible = (enct == 1)

        if name in ("", "_enct", "_b", "_benc", "_benc_h", "_benc_v"):
            if (enct == 0):
                be: int = self._benc.value
                enc = via.min_bottom_enclosure[b]

                vals["benc"] = enc.wide() if be == 0 else enc.tall()
            elif (enct == 1):
                be_h = self._benc_h.value
                be_v = self._benc_v.value

                vals["benc"] = _prp.Enclosure((be_h, be_v))

        if n_diff is not None:
            if name in ("", "_b"):
                if n_diff == b:
                    self._im.visible = True
                    self._imenc.visible = (enct == 0)
                    self._imenc_h.visible = (enct == 1)
                    self._imenc_v.visible = (enct == 1)
                else:
                    self._im.visible = False
                    self._imenc.visible = False
                    self._imenc_h.visible = False
                    self._imenc_v.visible = False

            if name in ("", "_im", "_imenc", "_imenc_h", "_imenc_v"):
                im = self._im.value

                if n_diff == b:
                    if (enct == 0):
                        imenc = self._imenc.value
                        enc = via.bottom[n_diff].min_implant_enclosure[im]

                        vals["imenc"] = enc.wide() if imenc == 0 else enc.tall()
                    elif (enct == 1):
                        imenc_h = self._imenc_h.value
                        imenc_v = self._imenc_v.value

                        vals["imenc"] = _prp.Enclosure((imenc_h, imenc_v))

        if name in ("", "_t"):
            enc = via.min_top_enclosure[t]
            self._tenc.visible = (enct == 0) and enc.is_assymetric
            self._tenc_h.visible = (enct == 1)
            self._tenc_v.visible = (enct == 1)

        if name in ("", "_tenc", "_tenc_h", "_tenc_v"):
            if (enct == 0):
                te: int = self._tenc.value
                enc = via.min_top_enclosure[t]

                vals["tenc"] = enc.wide() if te == 0 else enc.tall()
            elif (enct == 1):
                te_h = self._tenc_h.value
                te_v = self._tenc_v.value

                vals["tenc"] = _prp.Enclosure((te_h, te_v))

        if name in ("", "_mins"):
            self._space.visible = not self._mins.value

    def produce_impl(self):
        layoutfab = self._layoutfab
        gds_layers = self._gds_layers
        via = self._via
        vals = self._vals

        enct: int = self._enct
        rows: int = self._rows
        cols: int = self._cols
        b: int = self._b
        t: int = self._t
        im: int = self._im
        padw = self._padw
        padh = self._padh
        mins = self._mins
        space = self._space

        args = {}
        if len(via.bottom) > 1:
            args["bottom"] = via.bottom[b]
            if vals["n_diff"] is not None:
                if b == vals["n_diff"]:
                    diff = via.bottom[vals["n_diff"]]
                    assert isinstance(diff, _prm.WaferWire)
                    args["bottom_implant"] = diff.implant[im],
                    if enct in (0, 1):
                        args["bottom_implant_enclosure"] = vals["imenc"]
        if len(via.top) > 1:
            args["top"] = via.top[t]
        if enct in (0, 1):
            args.update({
                "rows": rows,
                "columns": cols,
                "bottom_enclosure": vals["benc"],
                "top_enclosure": vals["tenc"],
            })
        else:
            assert enct == 2
            args.update({
                "bottom_width": padw, "top_width": padw,
                "bottom_height": padh, "top_height": padh,
            })
        if not mins:
            args["space"] = space
        layout = layoutfab.layout_primitive(prim=via, **args)

        export2db(layout,
            gds_layers=gds_layers,
            pya_cell=self.cell,
            merge=True,
        )

class _PCellGenerator(PrimitiveDispatcher):
    def __init__(self, *, pya_lib: "PCellLibrary"):
        self._pya_lib = pya_lib
        self._pya_layout = pya_lib.layout()
        super().__init__()

    def __call__(self, prim: _prm.PrimitiveT) -> None:
        super().__call__(prim)

    def register(self, helper: pya.PCellDeclarationHelper, *, name: str):
        self._pya_layout.register_pcell(name, helper)

    # By default do nothing
    def _Primitive(self, prim: _prm.PrimitiveT) -> None:
        pass

    def Via(self, prim: _prm.Via):
        bottom_args: List[Tuple[str, Dict[str, Any]]] = []
        def add_bottom(*,
            bot: _prm.ViaBottom, enc: _prp.Enclosure,
            impl: Optional[_prm.Implant]=None, impl_enc: Optional[_prp.Enclosure]=None,
        ):
            b_args = {"bottom": bot} if len(prim.bottom) != 1 else {}
            if not enc.is_assymetric:
                args = (
                    (f"_{bot.name}", b_args),
                )
            else:
                args = (
                    (f"_{bot.name}$W", {**b_args, "bottom_enclosure": enc.wide()}),
                    (f"_{bot.name}$T", {**b_args, "bottom_enclosure": enc.tall()}),
                )
            if not impl:
                bottom_args.extend(args)
            else:
                assert isinstance(bot, _prm.WaferWire)
                assert impl_enc is not None
                if not impl_enc.is_assymetric:
                    bottom_args.extend((
                        (arg[0] + f"${impl.name}", {**arg[1], "bottom_implant": impl})
                        for arg in args
                    ))
                else:
                    bottom_args.extend((
                        (arg[0] + f"${impl.name}$W", {
                            **arg[1],
                            "bottom_implant": impl,
                            "bottom_implant_enclosure": impl_enc.wide(),
                        })
                        for arg in args
                    ))
                    bottom_args.extend((
                        (arg[0] + f"${impl.name}$T", {
                            **arg[1],
                            "bottom_implant": impl,
                            "bottom_implant_enclosure": impl_enc.tall(),
                        })
                        for arg in args
                    ))
        for i, bot in enumerate(prim.bottom):
            if not isinstance(bot, _prm.Resistor):
                enc = prim.min_bottom_enclosure[i]

                add_bottom(bot=bot, enc=enc)
                if isinstance(bot, _prm.WaferWire):
                    for i2, impl in enumerate(bot.implant):
                        if impl.type_ in (_prm.nImpl, _prm.pImpl):
                            add_bottom(
                                bot=bot, enc=enc,
                                impl=impl, impl_enc=bot.min_implant_enclosure[i2],
                            )

        top_args: List[Tuple[str, Dict[str, Any]]] = []
        def add_top(*, top: _prm.ViaTop, enc: _prp.Enclosure):
            t_args = {"top": top} if len(prim.top) != 1 else {}
            if not enc.is_assymetric:
                top_args.append((f"_{top.name}", t_args))
            else:
                top_args.extend((
                    (f"_{top.name}$W", {**t_args, "top_enclosure": enc.wide()}),
                    (f"_{top.name}$T", {**t_args, "top_enclosure": enc.tall()}),
                ))
        for i, top in enumerate(prim.top):
            if not isinstance(top, _prm.Resistor):
                enc = prim.min_top_enclosure[i]
                add_top(top=top, enc=enc)

        for barg, targ in product(bottom_args, top_args):
            name = prim.name + barg[0] + targ[0]
            cell = self._pya_layout.create_cell(name)
            lay = self._pya_lib._layoutfab.layout_primitive(
                prim=prim, **barg[1], **targ[1],
            )
            export2db(lay,
                gds_layers=self._pya_lib._gds_layers,
                pya_cell=cell,
                merge=True,
            )

        h = _ViaArray(
            layoutfab=self._pya_lib._layoutfab, gds_layers=self._pya_lib._gds_layers,
            via=prim,
        )
        self.register(h, name=f"{prim.name}$array")

    def MOSFET(self, prim: _prm.MOSFET) -> None:
        layoutfab = self._pya_lib._layoutfab
        gds_layers = self._pya_lib._gds_layers

        h = _MOSFET(
            layoutfab=layoutfab, gds_layers=gds_layers, mosfet=prim,
        )
        self.register(h, name=prim.name)

        h = _MOSFETFingers(
            layoutfab=layoutfab, gds_layers=gds_layers, mosfet=prim,
        )
        self.register(h, name=f"{prim.name}$fingers")


class PCellLibrary(pya.Library):
    def __init__(self, *,
        name: str, layoutfab: _lay.LayoutFactory, gds_layers: _typ.GDSLayerSpecDict,
    ):
        self._layoutfab = layoutfab
        self._gds_layers = gds_layers
        tech = layoutfab.tech
        self.description = f"PDKMaster based primitive library for {tech.name}"

        gen = _PCellGenerator(pya_lib=self)
        for prim in tech.primitives:
            gen(prim)

        self.register(name)
